import datetime

from django.forms import widgets
from django.utils.dates import MONTHS
from django.utils.translation import gettext_lazy as _

__all__ = (
    'Widget', 'MultiWidget',
    'SplitDateWidget', 'SplitHiddenDateWidget', 'SplitDateTimeWidget', 'SplitHiddenDateTimeWidget',

    'CheckboxInput',
    'CheckboxSelectMultiple', 'InlineCheckboxSelectMultiple', 'SeparatedCheckboxSelectMultiple',
    'RadioSelect', 'InlineRadioSelect', 'SeparatedRadioSelect',

    'TextInput', 'NumberInput', 'EmailInput', 'URLInput', 'PasswordInput', 'Textarea',
    'DateInput', 'DateTimeInput', 'TimeInput',
    'Select', 'NullBooleanSelect', 'SelectMultiple',

    'FileInput', 'ClearableFileInput',
    'SelectDateWidget',
)

group_template_names = (
    (widgets.Select, 'govuk_forms/field.html'),
    ((widgets.MultiWidget, widgets.ChoiceWidget), 'govuk_forms/field-fieldset.html'),
    (widgets.CheckboxInput, 'govuk_forms/field-no-label.html'),
    (widgets.Widget, 'govuk_forms/field.html'),
)


class Widget(widgets.Widget):
    input_classes = 'form-control'
    input_error_classes = 'form-control-error'

    def build_attrs(self, base_attrs, extra_attrs=None):
        attrs = super().build_attrs(base_attrs, extra_attrs=extra_attrs)
        css_classes = self.input_classes() if callable(self.input_classes) else self.input_classes
        attrs['class'] = ('%s %s' % (attrs.get('class', ''), css_classes)).strip()
        return attrs


class MultiWidget(widgets.MultiWidget, Widget):
    subwidget_group_classes = ()
    subwidget_label_classes = ()
    subwidget_labels = ()

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        iterator = zip(context['widget']['subwidgets'],
                       self.subwidget_group_classes,
                       self.subwidget_label_classes,
                       self.subwidget_labels)
        for subwidget, group_classes, label_classes, label in iterator:
            subwidget.update(
                group_classes=group_classes,
                label_classes=label_classes,
                label=label,
            )
        return context

    def decompress(self, value):
        raise NotImplementedError


class SplitDateWidget(MultiWidget):
    template_name = 'govuk_forms/widgets/split-date.html'
    subwidget_group_classes = ('form-group form-group-day',
                               'form-group form-group-month',
                               'form-group form-group-year')
    subwidget_label_classes = ('form-label', 'form-label', 'form-label')  # or form-label-bold
    subwidget_labels = (_('Day'), _('Month'), _('Year'))

    def __init__(self, attrs=None):
        date_widgets = (widgets.NumberInput(attrs=attrs),
                        widgets.NumberInput(attrs=attrs),
                        widgets.NumberInput(attrs=attrs),)
        super().__init__(date_widgets, attrs)

    def decompress(self, value):
        if value:
            return [value.day, value.month, value.year]
        return [None, None, None]


class SplitHiddenDateWidget(SplitDateWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for widget in self.widgets:
            widget.input_type = 'hidden'


class SplitDateTimeWidget(widgets.SplitDateTimeWidget, MultiWidget):
    template_name = 'govuk_forms/widgets/split-date.html'
    subwidget_group_classes = ('form-group form-group-date', 'form-group form-group-time')
    subwidget_label_classes = ('form-label', 'form-label')  # or form-label-bold
    subwidget_labels = (_('Date'), _('Time'))


class SplitHiddenDateTimeWidget(SplitDateTimeWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for widget in self.widgets:
            widget.input_type = 'hidden'


class CheckboxInput(widgets.CheckboxInput, Widget):
    template_name = 'govuk_forms/widgets/checkbox.html'
    inherit_label_from_field = True
    label = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.conditionally_revealed = {}

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        if self.label:
            context['widget']['label'] = self.label
        context['conditionally_revealed'] = self.conditionally_revealed.get(True)
        return context


class ChoiceWidget(widgets.ChoiceWidget, Widget):
    template_name = 'govuk_forms/widgets/multiple-select.html'
    option_template_name = 'govuk_forms/widgets/multiple-select-option.html'
    separate_last_option = False
    last_option_label = _('or')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.conditionally_revealed = {}

    @property
    def is_flat_list(self):
        return not any(isinstance(choice, (tuple, list)) for name, choice in self.choices)

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        context.update(
            is_flat_list=self.is_flat_list,
            separate_last_option=self.separate_last_option,
            last_option_label=self.last_option_label,
        )
        return context

    def create_option(self, name, value, label, selected, index, subindex=None, attrs=None):
        option = super().create_option(name, value, label, selected, index, subindex, attrs)
        if self.conditionally_revealed:
            option['conditionally_revealed'] = self.conditionally_revealed.get(value)
        return option


class CheckboxSelectMultiple(ChoiceWidget, widgets.CheckboxSelectMultiple):
    pass


class InlineCheckboxSelectMultiple(CheckboxSelectMultiple):
    field_group_classes = 'inline'


class SeparatedCheckboxSelectMultiple(CheckboxSelectMultiple):
    separate_last_option = True


class RadioSelect(ChoiceWidget, widgets.RadioSelect):
    pass


class InlineRadioSelect(RadioSelect):
    field_group_classes = 'inline'


class SeparatedRadioSelect(RadioSelect):
    separate_last_option = True


class TextInput(widgets.TextInput, Widget):
    pass


class NumberInput(widgets.NumberInput, Widget):
    input_classes = 'form-control form-control-1-8'


class EmailInput(widgets.EmailInput, Widget):
    pass


class URLInput(widgets.URLInput, Widget):
    input_classes = 'form-control form-control-2-3'


class PasswordInput(widgets.PasswordInput, Widget):
    input_classes = 'form-control form-control-1-4'


class Textarea(widgets.Textarea, Widget):
    def __init__(self, attrs=None):
        default_attrs = {'cols': '60', 'rows': '8'}
        if attrs:
            default_attrs.update(attrs)
        super().__init__(default_attrs)


class DateInput(widgets.DateInput, Widget):
    input_classes = 'form-control form-control-1-8'


class DateTimeInput(widgets.DateTimeInput, Widget):
    input_classes = 'form-control form-control-1-4'


class TimeInput(widgets.TimeInput, Widget):
    input_classes = 'form-control form-control-1-8'


class Select(widgets.Select, Widget):
    pass


class NullBooleanSelect(widgets.NullBooleanSelect, Widget):
    input_classes = 'form-control form-control-1-4'
    default_choices = (
        ('1', _('Not set')),
        ('2', _('Yes')),
        ('3', _('No')),
    )

    def __init__(self, attrs=None):
        super(widgets.Select, self).__init__(attrs, self.default_choices)


class SelectMultiple(widgets.SelectMultiple, Select):
    pass


class FileInput(widgets.FileInput, Widget):
    input_classes = ''
    input_error_classes = ''


class ClearableFileInput(widgets.ClearableFileInput, FileInput):
    pass


class SelectDateWidget(MultiWidget):
    template_name = 'govuk_forms/widgets/split-date.html'
    select_widget = Select
    none_value = (0, _('Not set'))
    subwidget_group_classes = ('form-group form-group-day-select',
                               'form-group form-group-month-select',
                               'form-group form-group-year-select')
    subwidget_label_classes = ('form-label', 'form-label', 'form-label')  # or form-label-bold
    subwidget_labels = (_('Day'), _('Month'), _('Year'))

    def __init__(self, attrs=None, years=None, months=None, empty_label=None):
        this_year = datetime.date.today().year
        self.years = [(i, i) for i in years or range(this_year, this_year + 10)]
        self.months = [(month_value, month_name) for month_value, month_name in (months or MONTHS).items()]
        self.days = [(i, i) for i in range(1, 32)]

        if isinstance(empty_label, (list, tuple)):
            self.year_none_value = (0, empty_label[0])
            self.month_none_value = (0, empty_label[1])
            self.day_none_value = (0, empty_label[2])
        else:
            none_value = (0, empty_label) if empty_label is not None else self.none_value
            self.year_none_value = none_value
            self.month_none_value = none_value
            self.day_none_value = none_value

        date_widgets = (self.select_widget(attrs=attrs, choices=self.days),
                        self.select_widget(attrs=attrs, choices=self.months),
                        self.select_widget(attrs=attrs, choices=self.years))
        super().__init__(date_widgets, attrs=attrs)

    def get_context(self, name, value, attrs):
        iterators = zip(
            self.widgets,
            (self.days, self.months, self.years),
            (self.day_none_value, self.month_none_value, self.year_none_value)
        )
        for widget, choices, none_value in iterators:
            widget.is_required = self.is_required
            widget.choices = choices if self.is_required else [none_value] + choices
        return super().get_context(name, value, attrs)

    def decompress(self, value):
        if value:
            return [value.day, value.month, value.year]
        return [None, None, None]


widget_replacements = {
    widgets.TextInput: (TextInput, ()),
    widgets.NumberInput: (NumberInput, ()),
    widgets.EmailInput: (EmailInput, ()),
    widgets.URLInput: (URLInput, ()),
    widgets.PasswordInput: (PasswordInput, ('render_value',)),
    widgets.Textarea: (Textarea, ()),
    widgets.DateInput: (DateInput, ('format',)),
    widgets.DateTimeInput: (DateTimeInput, ('format',)),
    widgets.TimeInput: (TimeInput, ('format',)),
    widgets.Select: (Select, ('choices',)),
    widgets.SelectMultiple: (SelectMultiple, ('choices',)),
    widgets.NullBooleanSelect: (NullBooleanSelect, ()),
    widgets.CheckboxInput: (CheckboxInput, ('check_test',)),
    widgets.CheckboxSelectMultiple: (CheckboxSelectMultiple, ('choices',)),
    widgets.RadioSelect: (RadioSelect, ('choices',)),
    widgets.SplitDateTimeWidget: (SplitDateTimeWidget, ()),  # TODO: migrate formats
    widgets.FileInput: (FileInput, ()),
    widgets.ClearableFileInput: (ClearableFileInput, ()),
    widgets.SelectDateWidget: (SelectDateWidget, ('years', 'months')),  # TODO: migrate empty values
}


def replace_widget(widget, replacements):
    replacement = replacements.get(widget.__class__)
    if not replacement:
        return widget
    replacement_widget, widget_args = replacement
    replacement_widget = replacement_widget(attrs=widget.attrs, **{
        widget_arg: getattr(widget, widget_arg)
        for widget_arg in widget_args
    })
    replacement_widget.is_required = widget.is_required
    replacement_widget.is_localized = widget.is_localized
    return replacement_widget



import distutils.log
import os

import setuptools


class SimpleCommand(setuptools.Command):
    user_options = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        cwd = os.getcwd()
        os.chdir(os.path.dirname(__file__))
        self.run_command()
        os.chdir(cwd)

    def run_command(self):
        raise NotImplementedError


class MakeMessages(SimpleCommand):
    description = 'update localisation messages files'

    def run_command(self):
        from django.core.management import call_command

        self.announce('Updating localisation message files', level=distutils.log.INFO)
        call_command('makemessages', all=True, no_wrap=True, keep_pot=True)


class CompileMessages(SimpleCommand):
    description = 'compile localisation messages files'

    def run_command(self):
        from django.core.management import call_command

        self.announce('Compiling localisation message files', level=distutils.log.INFO)
        call_command('compilemessages', fuzzy=False)


command_classes = {
    'makemessages': MakeMessages,
    'compilemessages': CompileMessages,
}



from collections import OrderedDict

from django import forms
from django.core.exceptions import ValidationError
from django.utils.crypto import get_random_string
from django.utils.encoding import force_text
from django.utils.html import conditional_escape, format_html_join
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from govuk_forms import widgets as govuk_widgets


class GOVUKForm(forms.Form):
    error_css_class = 'form-group-error'
    required_css_class = 'form-group-required'  # no default styling

    auto_replace_widgets = False
    group_template_names = govuk_widgets.group_template_names

    field_group_classes = 'form-group'
    field_group_panel_classes = 'panel panel-border-narrow js-hidden'
    field_label_classes = 'form-label'  # or form-label-bold
    field_help_classes = 'form-hint'

    error_summary_title = _('There are problems in the form')
    error_summary_template_name = 'govuk_forms/error-summary.html'

    submit_button_label = _('Submit')
    submit_button_template_name = 'govuk_forms/submit-button.html'

    reveal_conditionally = {}
    fieldsets = ()
    fieldset_template_name = 'govuk_forms/fieldset.html'

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('label_suffix', '')
        super().__init__(*args, **kwargs)

        if self.auto_replace_widgets:
            widget_replacements = govuk_widgets.widget_replacements
            if hasattr(self, 'widget_replacements'):
                widget_replacements = widget_replacements.copy().update(self.widget_replacements)
            for field in self.fields.values():
                field.widget = govuk_widgets.replace_widget(field.widget, widget_replacements)

        self.conditionally_revealed = {}
        for target_fields in self.reveal_conditionally.values():
            for target_field in target_fields.values():
                field = self.fields[target_field]
                self.conditionally_revealed[target_field] = {
                    'required': field.required,
                }
                field.required = False

    def __str__(self):
        return self.as_div()

    def clean(self):
        super().clean()
        for choice_field_name, details in self.reveal_conditionally.items():
            chosen_value = self.cleaned_data.get(choice_field_name, object())
            choice_field = self.fields[choice_field_name]
            multi_choice = isinstance(choice_field, forms.MultipleChoiceField)
            if not multi_choice:
                chosen_value = [chosen_value]
            for chosen_value in chosen_value:
                target_field_name = details.get(chosen_value)
                if not target_field_name:
                    continue
                conditionally_revealed = self.conditionally_revealed.get(target_field_name)
                if not conditionally_revealed.get('required'):
                    continue
                target_value = self.cleaned_data[target_field_name]
                target_field = self.fields[target_field_name]
                if target_value in target_field.empty_values:
                    self.add_error(target_field_name, ValidationError(target_field.error_messages['required'],
                                                                      code='required'))
        return self.cleaned_data

    def get_group_template_name(self, widget):
        for widget_classes, template_name in self.group_template_names:
            if isinstance(widget, widget_classes):
                return template_name
        raise ValueError('Cannot determine template name for widget %r' % widget)

    def as_div(self):
        rows = []
        included_fields = set()
        for field_name, conditionally_revealed in self.conditionally_revealed.items():
            included_fields.add(field_name)
            conditionally_revealed['html'] = self.render_field(field_name, self.fields[field_name], in_panel=True)
        for legend, field_names in self.fieldsets:
            included_fields.update(field_names)
            context = {
                'legend': legend,
                'contents': format_html_join('\n\n', '{}', (
                    (self.render_field(field_name, self.fields[field_name]),)
                    for field_name in field_names
                )),
            }
            rows.append((mark_safe(self.renderer.render(self.fieldset_template_name, context)),))
        rows.extend(
            (self.render_field(name, field),)
            for name, field in self.fields.items()
            if name not in included_fields
        )
        return format_html_join('\n\n', '{}', rows)

    def render_field(self, name, field, in_panel=False):
        bound_field = self[name]
        if bound_field.is_hidden:
            return bound_field

        widget = field.widget
        if hasattr(widget, 'conditionally_revealed'):
            widget.conditionally_revealed = {
                value: {
                    'bound_field': self[target_field],
                    'html': self.conditionally_revealed[target_field]['html'],
                }
                for value, target_field in self.reveal_conditionally.get(name, {}).items()
            }
        errors = [conditional_escape(error) for error in bound_field.errors]
        group_classes = self.field_group_panel_classes if in_panel else self.field_group_classes
        if hasattr(widget, 'field_group_classes'):
            group_classes = '%s %s' % (group_classes, widget.field_group_classes)
        group_classes = bound_field.css_classes(group_classes)
        label_classes = self.field_label_classes
        help_classes = self.field_help_classes
        if bound_field.label:
            label = conditional_escape(force_text(bound_field.label))
        else:
            label = ''
        if getattr(widget, 'inherit_label_from_field', False):
            widget.label = label
        if field.help_text:
            help_text = force_text(field.help_text)
        else:
            help_text = ''

        if errors:
            widget_classes = getattr(widget, 'input_error_classes', 'form-control-error')
        else:
            widget_classes = ''
        widget_attrs = {
            'class': widget_classes,
        }
        rendered_field = bound_field.as_widget(attrs=widget_attrs)
        if field.show_hidden_initial:
            rendered_field += bound_field.as_hidden(only_initial=True)

        field_context = {
            'bound_field': bound_field,
            'rendered_field': rendered_field,
            'errors': errors,
            'group_classes': group_classes.strip(),
            'label_classes': label_classes.strip(),
            'help_classes': help_classes.strip(),
            'label': label,
            'help_text': help_text,
        }
        group_template_name = self.get_group_template_name(widget)
        return mark_safe(self.renderer.render(group_template_name, field_context))

    def error_summary(self, error_summary_title=None):
        errors = self.errors
        if not errors:
            return ''

        non_field_errors = self.non_field_errors()
        field_errors = OrderedDict(
            (field, errors[field.name])
            for field in self
            if field.name in self.errors
        )
        context = {
            'error_summary_title': error_summary_title or self.error_summary_title,
            'random_string': get_random_string(4),
            'errors': errors,  # does not preserve field order
            'non_field_errors': non_field_errors,
            'field_errors': field_errors,
        }
        return mark_safe(self.renderer.render(self.error_summary_template_name, context))

    def submit_button(self, label=None):
        context = {'label': label or self.submit_button_label}
        return mark_safe(self.renderer.render(self.submit_button_template_name, context))
